@@ -65,6 +65,4 @@ from ti_python_module.ti_draw import *
 
 ########################################################################
 
-from ti_python_module.err import argument_error
-from ti_python_module.err import range_error
-from ti_python_module.err import type_error
\ No newline at end of file
+from ti_python_module.err import *
\ No newline at end of file
