@@ -1,4 +1,5 @@
 import ti_python_module.err as err
+from multimethod import multimethod
 """
 Class containing all TI-Hub commands. Used for debugging
 """
@@ -332,7 +333,7 @@ class brightness():
 
 ###########################################################################################
 
-def dht(port:str):
+class dht():
     """
     This device outputs a list consisting of the current temperature, humidity, type of sensor, and last cached read status.
 
@@ -342,19 +343,77 @@ def dht(port:str):
     Category: Hub / Add Input Device
 
 
-    Returns an array / list: [port]
+    Returns None
     """
 
-    err.type_error(str, "str", port)
-    err.argument_error(port, "IN 1", "IN 2")
+    def __init__(self, port:str) -> None:
 
-    
-    print("Setting port for input device 'dht' to '" + port + "'")
-    return [port]
+        """
+        This device outputs a list consisting of the current temperature, humidity, type of sensor, and last cached read status.
+
+
+        Available Ports: 'IN 1', 'IN 2'
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+
+        err.type_error(str, "str", port)
+        err.argument_error(port, "IN 1", "IN 2")
+
+        
+        print("Setting port for input device 'dht' to '" + port + "'")
+        return
+
+
+    def temp_measurement(self):
+        """
+        Measures the temperature from the dht sensor
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+
+        print("[dht] getting measured temperature")
+        return None
+
+    def humidity_measurement(self):
+        """
+        Measures the humidity from the dht sensor
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+
+        print("[dht] getting measured humidity")
+        return None
+
+    def t_h_measurement(self):
+        """
+        Measures the temperature and the humidity from the dht sensor
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+
+        print("[dht] getting measured temperature and humidity")
+        return None
       
 ###########################################################################################
 
-def ranger(port:str):
+class ranger():
     """
     This device outputs the current distance measurement from the specified ultrasonic ranger.
 
@@ -364,18 +423,43 @@ def ranger(port:str):
     Category: Hub / Add Input Device
 
 
-    Returns an array / list: [port]
+    Returns None
     """
 
+    def __init__(self, port:str) -> None:
+        """
+        This device outputs the current distance measurement from the specified ultrasonic ranger.
 
-    err.type_error(str, "str", port)
-    err.argument_error(port, "IN 1", "IN 2")
-    print("Setting port for input device 'ranger' to '" + port + "'")    
-    return [port]
+        Available Ports: 'IN 1', 'IN 2'
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+
+
+        err.type_error(str, "str", port)
+        err.argument_error(port, "IN 1", "IN 2")
+        print("Setting port for input device 'ranger' to '" + port + "'")    
+        return
+
+    def measurement(self):
+        """
+        Returns the measured values from the ultrasonic-ranger sensor in cm.
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+        print("[ultrasonic] getting measured distance in cm")
         
 ###########################################################################################
 
-def light_level(port:str):
+class light_level():
     """
     This device outputs the brightness level from the external light level (brightness) sensor.
 
@@ -386,19 +470,67 @@ def light_level(port:str):
     Category: Hub / Add Input Device
 
 
-    Returns an array / list: [port]
+    Returns None
     """
 
-    
-    err.type_error(str, "str", port)
-    err.argument_error(port, "IN 1", "IN 2", "IN 3")
+    def __init__(self, port:str) -> None:
+        """
+        This device outputs the brightness level from the external light level (brightness) sensor.
 
-    print("Setting port for input device 'light_level' to '" + port + "'")
-    return [port]
+
+        Available Ports: 'IN 1', 'IN 2', 'IN 3'
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
         
+        err.type_error(str, "str", port)
+        err.argument_error(port, "IN 1", "IN 2", "IN 3")
+
+        print("Setting port for input device 'light_level' to '" + port + "'")
+        return
+
+    def measurement(self):
+        """
+        Returns the measured light-level value.
+        
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+        print("[light_level] getting measured brightness")
+        return None
+
+    def range(self, min:int, max:int):
+        """
+        Reconfigures the range of the light_level sensor.
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns an array / list: [min, max]
+        """
+        err.type_error(int, "int", min)
+        err.type_error(int, "int", max)
+
+        err.range_error(0, None, min)
+        err.range_error(1, None, max)
+
+        if(min > max):
+            raise ValueError("ERROR: Parameter <min> has to be smaller then <max>!")
+
+        print("[light_level] setting range to '" + str(min) + "' to '" + str(max) + "'")
+        return [min, max]
+            
 ###########################################################################################
 
-def temperature(port:str):
+class temperature():
     """
     This device outputs the temperature reading from the external temperature sensor. The default configuration is to support the Seeed temperature sensor in 'IN 1', 'IN 2' or 'IN 3' ports. To use the TI LM19 Temperature sensor from the TI-Innovator™ Hub breadboard pack, edit the port to the BB pin in use and use an optional argument Example: mylm19=temperature("BB 5","TIANALOG")
     
@@ -411,18 +543,48 @@ def temperature(port:str):
     Category: Hub / Add Input Device
 
 
-    Returns an array / list: [port]
+    Returns None
     """
+    def __init__(self, port:str) -> None:
+        
+        """
+        This device outputs the temperature reading from the external temperature sensor. The default configuration is to support the Seeed temperature sensor in 'IN 1', 'IN 2' or 'IN 3' ports. To use the TI LM19 Temperature sensor from the TI-Innovator™ Hub breadboard pack, edit the port to the BB pin in use and use an optional argument Example: mylm19=temperature("BB 5","TIANALOG")
+        
+        Please Note, that this method will throw an ERROR with this library!
 
-    err.type_error(str, "str", port)
-    err.argument_error(port, "IN 1", "IN 2", "IN 3")
 
-    print("Setting port for input device 'temperature' to '" + port + "'")
-    return [port]
+        Available Ports: 'IN 1', 'IN 2', 'IN 3'
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+
+        err.type_error(str, "str", port)
+        err.argument_error(port, "IN 1", "IN 2", "IN 3")
+
+        print("Setting port for input device 'temperature' to '" + port + "'")
+        return
+
+
+    def measurement(self):
+        """
+        Gets the measured temperature.
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+        print("[temperature] getting measued temperature")
+        return None
         
 ###########################################################################################
 
-def moisture(port:str):
+class moisture():
     """
     This device outputs the moisture sensor reading.
 
@@ -433,18 +595,66 @@ def moisture(port:str):
     Category: Hub / Add Input Device
 
 
-    Returns an array / list: [port]
+    Returns None
     """
+    def __init__(self, port:str) -> None:
+        """
+        This device outputs the moisture sensor reading.
 
-    err.type_error(str, "str", port)
-    err.argument_error(port, "IN 1", "IN 2", "IN 3")
 
-    print("Setting port for input device 'moisture' to '" + port + "'")
-    return [port]
+        Available Ports: 'IN 1', 'IN 2', 'IN 3'
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+
+        err.type_error(str, "str", port)
+        err.argument_error(port, "IN 1", "IN 2", "IN 3")
+
+        print("Setting port for input device 'moisture' to '" + port + "'")
+        return
+
+    def measurement(self):
+        """
+        Returns the measured moisture.
+        
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+        print("[moisture] getting measured moisture")
+        return None
+
+    def range(self, min:int, max:int):
+        """
+        Reconfigures the range of the moisture sensor.
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns an array / list: [min, max]
+        """
+        err.type_error(int, "int", min)
+        err.type_error(int, "int", max)
+
+        err.range_error(0, None, min)
+        err.range_error(1, None, max)
+
+        if(min > max):
+            raise ValueError("ERROR: Parameter <min> has to be smaller then <max>!")
+
+        print("[moisture] setting range to '" + str(min) + "' to '" + str(max) + "'")
+        return [min, max]
         
 ###########################################################################################
 
-def magnetic(port:str):
+class magnetic():
     """
     This device detects the presence of a magnetic field. The threshold value to determine the presence of the field is set through the trigger() function. The default value of the threshold is 150.
 
@@ -456,18 +666,74 @@ def magnetic(port:str):
     Category: Hub / Add Input Device
 
 
-    Returns an array / list: [port]
+    Returns None
     """
+    def __init__(self, port:str) -> None:
+        """
+        This device detects the presence of a magnetic field. The threshold value to determine the presence of the field is set through the trigger() function. The default value of the threshold is 150.
 
-    err.type_error(str, "str", port)
-    err.argument_error(port, "IN 1", "IN 2", "IN 3")
 
-    print("Setting port for input device 'magnetic' to '" + port + "'")
-    return [port]
-     
+
+        Available Ports: 'IN 1', 'IN 2', 'IN 3'
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+
+        err.type_error(str, "str", port)
+        err.argument_error(port, "IN 1", "IN 2", "IN 3")
+
+        print("Setting port for input device 'magnetic' to '" + port + "'")
+        return
+
+    def measurement(self):
+        """
+        Returns the measured magnetic value
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+        print("[magnetic] measuring magnetic value")
+        return None
+
+    def magnet_close(self):
+        """
+        Says, whether a magnet is close, by checking if the magnetic value is over the threshold. This can be changed using the threshold function.
+        
+        
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+        print("[magnetic] checking if magnet is close (magnetic value is over threshold)")
+        return None
+
+    def trigger(self, threshold:int):
+        """
+        Sets the threshold of the magnetic value, at which the magnet_close functions triggers. This ranges from 0 to 16383. The default is 150.
+        
+        
+        Category: Hub / Add Input Device
+
+
+        Returns an array / list: [threshold]
+        """
+        err.type_error(int, "int", threshold)
+        err.range_error(0, 16383)
+
+        print("[magnetic] setting threshold to '" + str(threshold) + "'")
+        return [threshold]
+
 ###########################################################################################
 
-def vernier(port:str, sensor_type:str):
+class vernier():
     """
     This device reads the value from the vernier analog sensor specified by the <sensor_type>.\n
     Supported Features:\n
@@ -490,19 +756,81 @@ def vernier(port:str, sensor_type:str):
 
     Returns an array / list: [port]
     """
+    def __init__(self, port:str, sensor_type:str) -> None:
+        """
+        This device reads the value from the vernier analog sensor specified by the <sensor_type>.\n
+        Supported Features:\n
+        • temperature - Stainless Steel Temperature sensor.\n
+        • lightlevel - TI Light level sensor.\n
+        • pressure - Original gas pressure sensor\n
+        • pressure2 - Newer gas pressure sensor.\n
+        • pH - pH sensor.\n
+        • force10 - ±10 N setting, Dual Force Sensor.\n
+        • force50 - ±50 N setting, Dual Force Sensor.\n
+        • accelerometer - Low-G Accelerometer.\n
+        • generic - Allows setting of other sensors not supported directly above, and use of the calibrate() API above to set equation coefficients\n<
 
-    err.type_error(str, "str", port)
-    err.type_error(str, "str", sensor_type)
-    err.argument_error(port, "IN 1", "IN 2", "IN 3")
-    err.argument_error(sensor_type, "temperature", "lightlevel", "pressure", "pressure2", "pH", "force10", "force50", "accelerometer", "generic")
 
-    print("Setting port for input device 'vernier' to '" + port + "' with sensor type '" + sensor_type + "'")
-    return [port]
+        Available Ports: 'IN 1', 'IN 2', 'IN 3'
 
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+
+        
+
+        err.type_error(str, "str", port)
+        err.type_error(str, "str", sensor_type)
+        err.argument_error(port, "IN 1", "IN 2", "IN 3")
+        err.argument_error(sensor_type, "temperature", "lightlevel", "pressure", "pressure2", "pH", "force10", "force50", "accelerometer", "generic")
+
+        print("Setting port for input device 'vernier' to '" + port + "' with sensor type '" + sensor_type + "'")
+        return
+
+    def measurement(self):
+        """
+        Outputs the measured value of the set sensor type.
+        
+        
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+        print("[vernier] measuring value from set sensor type")
+        return None
+
+    def calibrate(a, b, c = None, d = None):
+        """
+        Calibrates the sensor.
+        
+        
+        Use type 1: a, b ==> linear: ax + b\n
+        Use type 2: a, b, c, d
+        
+        
+        Category: Hub / Add Input Device
+
+
+        Returns an array / list: [a, b, c, d]
+        """
+        if(c == None and d == None):
+            print("[vernier] calibrating linearly with '" + str(a) + "' and '" + str(b) + "'")
+        elif(c != None and d == None):
+            raise ValueError("ERROR: If you specify 'c' you also need to specify 'd'")
+        else:
+            print("[vernier] calibrating with '" + str(a) + "', '" + str(b) + "', '" + str(c) + "' and '" + str(d) + "'")
+
+        return [a, b, c, d]
+
+        
         
 ###########################################################################################
 
-def analog_in(port:str):
+class analog_in():
 
     """
     This device supports the use of analog input generic devices.
@@ -515,41 +843,68 @@ def analog_in(port:str):
     Category: Hub / Add Input Device
 
 
-    Returns an array / list: [port]
+    Returns None
     """
+    def __init__(self, port:str) -> None:
+        """
+        This device supports the use of analog input generic devices.
 
-    err.type_error(str, "str", port)
-    err.argument_error(port, "IN 1", "IN 2", "IN 3", "BB 5", "BB 6", "BB 7")
 
-    print("Setting port for input device 'analog_in' to '" + port + "'")
-    return [port]
+
+        Available Ports: 'IN 1', 'IN 2', 'IN 3', 'BB 5', 'BB 6', 'BB 7'
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+
+        err.type_error(str, "str", port)
+        err.argument_error(port, "IN 1", "IN 2", "IN 3", "BB 5", "BB 6", "BB 7")
+
+        print("Setting port for input device 'analog_in' to '" + port + "'")
+        return
+
+    def measurement(self):
+        """
+        Outputs the measured value from the sensor set to 'analog_in'.
+        
+        
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+        print("[analog_in] measuring sensor value")
+        return None
+
+    def range(self, min:int, max:int):
+        """
+        Reconfigures the range of the sensor set to 'analog_in'.
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns an array / list: [min, max]
+        """
+        err.type_error(int, "int", min)
+        err.type_error(int, "int", max)
+
+        err.range_error(0, None, min)
+        err.range_error(1, None, max)
+
+        if(min > max):
+            raise ValueError("ERROR: Parameter <min> has to be smaller then <max>!")
+
+        print("[analog_in] setting range to '" + str(min) + "' to '" + str(max) + "'")
+        return [min, max]
+        
         
 ###########################################################################################
 
-def digital(port:str):
-    """
-    This device outputs the current state of the digital pin connected to the 'DIGITAL' object, or the cached state of the digital output value last 'SET' to the object.
-
-
-
-    Available Ports: 'IN 1', 'IN 2', 'IN 3', 'BB 1', 'BB 2', 'BB 3', 'BB 4', 'BB 5', 'BB 6', 'BB 7', 'BB 8', 'BB 9', 'BB 10'
-
-
-    Category: Hub / Add Input Device
-
-
-    Returns an array / list: [port]
-    """
-
-    err.type_error(str, "str", port)
-    err.argument_error(port, "IN 1", "IN 2", "IN 3", "BB 1", "BB 2", "BB 3", "BB 4", "BB 5", "BB 6", "BB 7", "BB 8", "BB 9", "BB 10")
-
-    print("Setting port for input device 'digital' to '" + port + "'")
-    return [port]
-        
-###########################################################################################
-
-def potentiometer(port:str):
+class potentiometer():
     """
     This device supports a potentiometer sensor. The range of the sensor can be changed by the range() function.
 
@@ -561,18 +916,67 @@ def potentiometer(port:str):
     Category: Hub / Add Input Device
 
 
-    Returns an array / list: [port]
+    Returns None
     """
+    def __init__(self, port:str) -> None:
+        """
+        This device supports a potentiometer sensor. The range of the sensor can be changed by the range() function.
 
-    err.type_error(str, "str", port)
-    err.argument_error(port, "IN 1", "IN 2", "IN 3", "BB 5", "BB 6", "BB 7")
 
-    print("Setting port for input device 'potentiometer' to '" + port + "'")
-    return [port]
+
+        Available Ports: 'IN 1', 'IN 2', 'IN 3', 'BB 5', 'BB 6', BB 7'
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+
+        err.type_error(str, "str", port)
+        err.argument_error(port, "IN 1", "IN 2", "IN 3", "BB 5", "BB 6", "BB 7")
+
+        print("Setting port for input device 'potentiometer' to '" + port + "'")
+        return
         
+
+    def measurement(self):
+        """
+        Outputs the measured value from the 'potentiometer'.
+        
+        
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+        print("[potentiometer] measuring sensor value")
+        return None
+
+    def range(self, min:int, max:int):
+        """
+        Reconfigures the range of the 'potentiometer'.
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns an array / list: [min, max]
+        """
+        err.type_error(int, "int", min)
+        err.type_error(int, "int", max)
+
+        err.range_error(0, None, min)
+        err.range_error(1, None, max)
+
+        if(min > max):
+            raise ValueError("ERROR: Parameter <min> has to be smaller then <max>!")
+
+        print("[potentiometer] setting range to '" + str(min) + "' to '" + str(max) + "'")
+        return [min, max]
 ###########################################################################################
 
-def thermistor(port:str):
+class thermistor():
     """
     This device reads thermistor sensors. The default coefficients are designed to match the thermistor included in the Breadboard Pack of the TI-Innovator™ Hub, when used with a 10KΩ fixed resistor. A new set of calibration coefficients and reference resistance for the thermistor can be configured using the calibrate() function.
 
@@ -584,18 +988,58 @@ def thermistor(port:str):
     Category: Hub / Add Input Device
 
 
-    Returns an array / list: [port]
+    Returns None
     """
+    def __init__(self, port:str) -> None:
+        """
+        This device reads thermistor sensors. The default coefficients are designed to match the thermistor included in the Breadboard Pack of the TI-Innovator™ Hub, when used with a 10KΩ fixed resistor. A new set of calibration coefficients and reference resistance for the thermistor can be configured using the calibrate() function.
 
-    err.type_error(str, "str", port)
-    err.argument_error(port, "IN 1", "IN 2", "IN 3", "BB 5", "BB 6", "BB 7")
 
-    print("Setting port for input device 'thermistor' to '" + port + "'")
-    return [port]
+
+        Available Ports: 'IN 1', 'IN 2', 'IN 3', 'BB 5', 'BB 6', BB 7'
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+
+        err.type_error(str, "str", port)
+        err.argument_error(port, "IN 1", "IN 2", "IN 3", "BB 5", "BB 6", "BB 7")
+
+        print("Setting port for input device 'thermistor' to '" + port + "'")
+        return
+
+    def measurement(self):
+        """
+        Outputs the measured value from the 'thermistor'.
+        
+        
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+        print("[thermistor] measuring sensor value")
+        return None
+
+    def calibrate(self, c1, c2, c3, r):
+        """
+        Calibrates the 'potentiomter'.
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns an array / list: [c1, c2, c3, r]
+        """
+        print("[thermistor] calibrating with '" + str(c1) + "', '" +str(c2) + "', '" +str(c3) + "' and '" + str(r))
+        return [c1, c2, c3, r]
         
 ###########################################################################################
 
-def loudness(port:str):
+class loudness():
     """
     This device supports sound loudness sensors.
 
@@ -606,15 +1050,63 @@ def loudness(port:str):
     Category: Hub / Add Input Device
 
 
-    Returns an array / list: [port]
+    Returns None
     """
+    def __init__(self, port:str) -> None:
+        """
+        This device supports sound loudness sensors.
 
-    err.type_error(str, "str", port)
-    err.argument_error(port, "IN 1", "IN 2", "IN 3")
 
-    print("Setting port for input device 'loudness' to '" + port + "'")
-    return [port]
+        Available Ports: 'IN 1', 'IN 2', 'IN 3'
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+
+        err.type_error(str, "str", port)
+        err.argument_error(port, "IN 1", "IN 2", "IN 3")
+
+        print("Setting port for input device 'loudness' to '" + port + "'")
+        return
         
+    def measurement(self):
+        """
+        Outputs the measured value from the 'loudness'.
+        
+        
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+        print("[loudness] measuring sensor value")
+        return None
+
+
+    def range(self, min:int, max:int):
+        """
+        Reconfigures the range of the 'loudness'.
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns an array / list: [min, max]
+        """
+        err.type_error(int, "int", min)
+        err.type_error(int, "int", max)
+
+        err.range_error(0, None, min)
+        err.range_error(1, None, max)
+
+        if(min > max):
+            raise ValueError("ERROR: Parameter <min> has to be smaller then <max>!")
+
+        print("[loudness] setting range to '" + str(min) + "' to '" + str(max) + "'")
+        return [min, max]
 ###########################################################################################
 class colour_input():
 
@@ -734,63 +1226,8 @@ class colour_input():
         return None
 
 ###########################################################################################
-class bb_port():
-    """
-    This device provides support for using all 10 BB port pins as a combined digital input/output port. The initialization functions have an optional "mask" parameter that allows the use of the subset of the 10 pins.\n
-    • read_port(): Reads the current values on the input pins of the BB port.\n
-    • write_port(value): Sets the output pin values to the specified value, where value is between 0 and 1023. Note that the value is also adjusted against the mask value in the var=bbport(mask) operation, if a mask was provided.
 
-
-    Category: Hub / Add Input Device
-
-
-    Returns None
-    """
-
-    def __init__(self, mask:int) -> None:
-        """
-        This device provides support for using all 10 BB port pins as a combined digital input/output port. The initialization functions have an optional "mask" parameter that allows the use of the subset of the 10 pins.\n
-        • read_port(): Reads the current values on the input pins of the BB port.\n
-        • write_port(value): Sets the output pin values to the specified value, where value is between 0 and 1023. Note that the value is also adjusted against the mask value in the var=bbport(mask) operation, if a mask was provided.
-
-
-        Category: Hub / Add Input Device
-
-
-        Returns None
-        """
-
-        err.type_error(int, "int", mask)
-        err.range_error(-1, 1024, mask)
-
-
-        print("Setting mask for input device 'bb_port' to '" + str(mask) + "'")
-        return 
-
-
-            
-
-    def read_port(self):
-        """
-        Reads the current values on the input pins of the BB port.\n
-        """
-        print("[bb_port] reading ports")
-        return
-    
-    def write_port(self ,value:int):
-        """
-        Sets the output pin values to the specified value, where value is between 0 and 1023. Note that the value is also adjusted against the mask value in the var=bbport(mask) operation, if a mask was provided.
-
-
-        Returns an array / list: [value]
-        """
-        err.type_error(int, "int", value)
-        print("[bb_port] writing value " + str(value))
-        return [value]
-
-###########################################################################################
-
-def hub_time():
+class hub_time():
     """
     This device provides access to the internal millisecond timer of the HUB.
 
@@ -800,8 +1237,43 @@ def hub_time():
 
     Returns None
     """
-    print("Fetching millisecond timer of HUB")
-    return None
+    def __init__(self) -> None:
+        """
+        This device provides access to the internal millisecond timer of the HUB.
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+        return
+
+    def measurement(self):
+        """
+        Outputs the measured value from the 'hub_time'.
+        
+        
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+        print("[hub_time] Fetching millisecond timer of HUB")
+        return None
+
+    def reset_time(self):
+        """
+        Resets the HUB-intern millisecond counter.
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+        print("[hub_time] resetting hub timer")
+        return None
 
 ###########################################################################################
 class rgb_array():
@@ -1449,5 +1921,160 @@ class squarewave():
         print("[Square Wave] Turning squarewave off")
         return None
 
-##digital - out <> just use as a class with __init__ method as default. add additional submethods later
-##bb_port - out <> ''
+
+class digital():
+    """
+    INPUT: This device outputs the current state of the digital pin connected to the 'DIGITAL' object, or the cached state of the digital output value last 'SET' to the object.\n
+    OUTPUT: Interface for controlling a digital output
+
+
+
+    INPUT: Available Ports: 'IN 1', 'IN 2', 'IN 3', 'BB 1', 'BB 2', 'BB 3', 'BB 4', 'BB 5', 'BB 6', 'BB 7', 'BB 8', 'BB 9', 'BB 10'\n
+    INPUT: Available Ports: 'OUT 1', 'OUT 2', 'OUT 3', 'BB 1', 'BB 2', 'BB 3', 'BB 4', 'BB 5', 'BB 6', 'BB 7', 'BB 8', 'BB 9', 'BB 10'
+
+
+    Category: Hub / Add Input Device\n
+    Category: Hub / Add Output Device
+
+
+    Returns None
+    """
+
+    def __init__(self, port:str) -> None:
+        """
+        INPUT: This device outputs the current state of the digital pin connected to the 'DIGITAL' object, or the cached state of the digital output value last 'SET' to the object.\n
+        OUTPUT: Interface for controlling a digital output
+
+
+
+        INPUT: Available Ports: 'IN 1', 'IN 2', 'IN 3', 'BB 1', 'BB 2', 'BB 3', 'BB 4', 'BB 5', 'BB 6', 'BB 7', 'BB 8', 'BB 9', 'BB 10'\n
+        INPUT: Available Ports: 'OUT 1', 'OUT 2', 'OUT 3', 'BB 1', 'BB 2', 'BB 3', 'BB 4', 'BB 5', 'BB 6', 'BB 7', 'BB 8', 'BB 9', 'BB 10'
+
+
+        Category: Hub / Add Input Device\n
+        Category: Hub / Add Output Device
+
+
+        Returns None
+        """
+
+        err.type_error(str, "str", port)
+        err.argument_error(port, "OUT 1", "OUT 2", "OUT 3", "IN 1", "IN 2", "IN 3", "BB 1", "BB 2", "BB 3", "BB 4", "BB 5", "BB 6", "BB 7", "BB 8", "BB 9", "BB 10")
+
+        print("Setting port for input / output device 'digital' to '" + port + "'")
+
+
+    def measurement(self):
+        """
+        Returns the value of the digital input device
+
+
+        Category: Hub / Add Input Device
+
+
+        Returns None
+        """
+        print("Measuring value of digital inout device")
+
+
+    def set(self ,value:int):
+        """
+        Sets the digital output to the value specified by "value" (0 or 1).
+
+
+        Category: Hub / Add Output Device
+
+
+        Returns an array / list: [value]
+        """
+
+        err.type_error(int, "int", value)
+        err.range_error(0, 1, value)
+
+        print("Setting digital outout to '" + str(value) + "'")
+        return [value]
+
+
+    def on(self):
+        """
+        Sets the state of the digital output to high (1).
+
+
+        Category: Hub / Add Output Device
+
+
+        Returns None
+        """
+        print("Setting state of digital output device to 'on' (1)")
+        return None
+
+    def off(self):
+        """
+        Sets the state of the digital output to high (1).
+
+
+        Category: Hub / Add Output Device
+
+
+        Returns None
+        """
+        print("Setting state of digital output device to 'off' (0)")
+        return None
+
+###########################################################################################
+
+class bb_port():
+    """
+    This device provides support for using all 10 BB port pins as a combined digital input/output port. The initialization functions have an optional "mask" parameter that allows the use of the subset of the 10 pins.\n
+    • read_port(): Reads the current values on the input pins of the BB port.\n
+    • write_port(value): Sets the output pin values to the specified value, where value is between 0 and 1023. Note that the value is also adjusted against the mask value in the var=bbport(mask) operation, if a mask was provided.
+
+
+    Category: Hub / Add Input Device\n
+    Category: Hub / Add Output Device
+
+
+    Returns None
+    """
+
+    def __init__(self, mask:int) -> None:
+        """
+        This device provides support for using all 10 BB port pins as a combined digital input/output port. The initialization functions have an optional "mask" parameter that allows the use of the subset of the 10 pins.\n
+        • read_port(): Reads the current values on the input pins of the BB port.\n
+        • write_port(value): Sets the output pin values to the specified value, where value is between 0 and 1023. Note that the value is also adjusted against the mask value in the var=bbport(mask) operation, if a mask was provided.
+
+
+        Category: Hub / Add Input Device\n
+        Category: Hub / Add Output Device
+
+
+        Returns None
+        """
+
+        err.type_error(int, "int", mask)
+        err.range_error(-1, 1024, mask)
+
+
+        print("Setting mask for input / output device 'bb_port' to '" + str(mask) + "'")
+        return 
+
+
+            
+
+    def read_port(self):
+        """
+        Reads the current values on the input pins of the BB port.\n
+        """
+        print("[bb_port] reading ports")
+        return
+    
+    def write_port(self ,value:int):
+        """
+        Sets the output pin values to the specified value, where value is between 0 and 1023. Note that the value is also adjusted against the mask value in the var=bbport(mask) operation, if a mask was provided.
+
+
+        Returns an array / list: [value]
+        """
+        err.type_error(int, "int", value)
+        print("[bb_port] writing value " + str(value))
+        return [value]
\ No newline at end of file
